#include"classwriter.h"
#include<ctype.h>
#include"output.h"
#include"typename2id.h"
#include"classname2names.h"
// Interface
extern "C" const char *get_semantik_name();

void CommonContentElement::addComment(const char *text,size_t len)
{
    for(const char *c=text;c<text+len;c++)
    {
	comment_ << *c;
    }
}

void CommonContentElement::finish()
{
    std::string comment = comment_.str();
    if(comment.size()>0)
    {
	cw_->decl() << indent(cw_->level()+1) 
		    << "/* " << comment_.str() << " */" << std::endl;
    }
    cw_->decl() << decl_.str();
    comment_.str("");
    decl_.str("");
}

void CommonContentElement::checkfinished()
{
    if(!finished_) finish();
}

ClassWriter::ClassWriter(Scope *scope,const char *name,const char *id)
    :finished_(false),namespace_(0),name_(name),level_(0),
     numbersuperclasses_(0),iteratorheaderincl_(false),contentele_(this)
{
    Classname2names cnn(name);
    const char *namesp = (scope ? scope->Name().c_str() : 0);
    if(namesp!=0)
    {
	namespace_=strdup(namesp);
	level_++;
    }
    
    header_.open(cnn.headerfilename(),std::ios_base::out);
    header_ << "#ifndef " << cnn.includestopper() << std::endl;
    header_ << "#define " << cnn.includestopper() << std::endl;
    includes_ << "#include<RAMS/GrowingBuffer.h>" << std::endl;
    includes_ << "#include\"PrimitivTypes.h\"" << std::endl; 
    classhead_ << indent(level_) << "class " << name;
    

    body_.open(cnn.bodyfilename(),std::ios_base::out);
    body_ << "#include\"" << cnn.headerfilename() << "\"\n" << std::endl;
    body_ << "#include<ramsmof/Resolver.h>" << std::endl;
    const char *sname = get_semantik_name();
    if(sname)
    {
	body_ << "#include\"" << sname << "_Semantik.h\"" << std::endl;
    }

    if(namesp !=0)
    {
	body_ << "using namespace " << namesp << ";\n" << std::endl;
    }
   
    setsegbody_ << "void " << name_ << "::_setsegandrep(RAMS::GrowingBuffer *place,RAMS::octet_t *rep)\n{\n"
	  << indent(1) << "place_=place;\n"
	  << indent(1) << "rep_=(" << name_ << "::Layout *)rep;\n";
   
    //??? what to do if place->freespace() < sizeof(Layout) ???
    initfunc_ << "void " << name_ << "::_init()\n{" << std::endl;
}

std::string ClassWriter::Name()
{
    return name_;
}

void ClassWriter::addSuperClass(const char *name)
{
    includes_ << "#include\"" << name << ".h\"" << std::endl;
    classhead_ << (numbersuperclasses_ ? ", " : " : ")
		 << "public virtual " << name;
    layouthead_ << indent(level_+2) << name << "::Layout rep_" << name << ";\n";
    constructorbody_ << indent(1) << name 
		     << "::_setsegandrep(place_,(RAMS::octet_t *)&(rep_->rep_" << name 
		     << "));" << std::endl;
    setsegbody_      << indent(1) << name 
		     << "::_setsegandrep(place_,(RAMS::octet_t *)&(rep_->rep_" << name 
		     << "));" << std::endl;
    initfunc_   << indent(1) << name << "::_init();" << std::endl; 
    numbersuperclasses_++;
}

ElementGenerator *ClassWriter::addProperty(const char *name,const char *id,
			      const char *typ,bool type_is_elementar,
			      const char *def)
{
    contentele_.checkfinished();
    classpublic_ << indent(level_+1) << "/* id=" << id << "*/" << std::endl;
    layout_ << indent(level_+2) << "/* id=" << id << "*/" << std::endl;
    body_ << "/* id=" << id << "*/" << std::endl;

    if(!type_is_elementar)
    {
	includes_ << "#include\"" << typ << ".h\"" << std::endl;
	//* We don't know the exact type: it may be a subtype */
	layout_ << indent(level_+2) << "::RAMS::MOF::typeid_t typid_"
		<< name << "_;" << std::endl;
	layout_ << indent(level_+2) << "off_t " << name << ";"
		<< std::endl;
	contentele_.decl() << indent(level_+1) << typ << " *" << name << "();"
		     << std::endl;
	contentele_.decl() << indent(level_+1) << "void " << name << "(" 
			   << typ << " *valueref);";

	initfunc_ << indent(1) << "rep_->typid_" << name << "_ = 0;\n"
		  << indent(1) << "rep_->" << name << " = RAMS::RNIL;"
		  << std::endl;

	body_ << typ << " *" << name_ << "::" << name << "()\n{\n"
	      << indent(1) << "if(rep_->" << name << " == RAMS::RNIL) "
	      << "return 0;\n"
	      << indent(1) << typ << " *ret=dynamic_cast<" << typ
	      << " *>(resolve(rep_->typid_" << name << "_));\n"
	      << indent(1) << "ret->_setsegandrep(place_,place_->ptr(rep_->" 
	      << name << "));\n"
	      << indent(1) << "return ret;\n"
	      << "}\n" << std::endl;
	body_ << "void " << name_ << "::" << name << "(" 
	      << typ << " *valueref)\n{\n"
	      << indent(1) << "rep_->typid_" << name << "_=valueref->_typeid();\n"
	      << indent(1) << "place_->wrptr_align(2);\n"
	      << indent(1) << "rep_->" << name << "=place_->wr_rel();\n"
	      << indent(1) << "valueref->_setsegandrep(place_,place_->wr_ptr(valueref->_layout_size()));\n"
	      << indent(1) << "valueref->_init();\n"
	      << "}\n" << std::endl;
    }
    else if(strcmp(typ,"String")==0)
    {
	contentele_.decl() << indent(level_+1) << "const RAMS::String &"
		     << name << "();" << std::endl;
	contentele_.decl() << indent(level_+1) << "void " << name
		     << "(const char *value);" << std::endl;
	contentele_.decl() << indent(level_+1) << "void " << name
		     << "(const char *value,size_t len);\n" << std::endl;
	layout_ << indent(level_+2) << "off_t " << name << ";" << std::endl;
	body_ << "const RAMS::String & " << name_ << "::" << name << "()\n"
	      << "{\n"
	      << indent(1) << "return *((RAMS::String *)((RAMS::octet_t *)(place_->begin())+rep_->"
	      << name << "));\n"
	      << "}\n\n" << std::endl;
	body_ << "void " << name_ << "::" << name << "(const char *value)\n" 
	      << "{\n";	
	body_ << indent(1) << "RAMS::String *z;\n"
	      << indent(1) << "const char *p=value;\n"
	      << indent(1) << "place_->wrptr_align(2);\n"
	      << indent(1) << "rep_->" << name << " = place_->wr_rel();\n"
	      << indent(1) << "z=(RAMS::String *)(place_->wr_ptr(sizeof(size_t)));\n"
	      << indent(1) << "z->size=0;\n"
	      << indent(1) << "while(*p!='\\0')\n"
	      << indent(1) << "{\n"
	      << indent(2) << "*(place_->wr_ptr(1))=*p;\n"
	      << indent(2) << "p++; (z->size)++;\n"
	      << indent(1) << "}\n"
	      << "}\n" << std::endl;
	body_ << "void " << name_ << "::" << name << "(const char *value,size_t len)\n{\n"
              << indent(1) << "RAMS::String *z;\n"
	      << indent(1) << "place_->wrptr_align(2);\n"
	      << indent(1) << "rep_->" << name << " = place_->wr_rel();\n"
              << indent(1) << "z=(RAMS::String *)(place_->wr_ptr(sizeof(size_t)+len));\n"
	      << indent(1) << "z->size=len;\n"
	      << indent(1) << "memcpy(z->mem,value,len);\n"
	      << "}\n";
	initfunc_ << indent(1) << "rep_->" << name << " = RAMS::RNIL;"
		  << std::endl;
	if(def)
	{
	    initfunc_ << indent(1) << name << "(\"" << def
		      << "\");" << std::endl;
	}
    }
    else
    {
        contentele_.decl() << indent(level_+1) << typ << " " 
			   << name << "();" << std::endl;
	contentele_.decl() << indent(level_+1) << "void " << name << "("
			   << typ << " value);" << std::endl;
    
	layout_ << indent(level_+2) << typ << " " << name << ";" << std::endl;
	body_ << typ << " " << name_ << "::" << name << "()\n"
	      << "{\n"
	      << indent(1) << "return rep_->" << name << ";\n"
	      << "}\n" << std::endl;
	body_ << "void " << name_ << "::" << name << "(" << typ << " value)\n" 
	      << "{\n"
	      << indent(1) << "rep_->" << name << " = value;\n"
	      << "}\n" << std::endl;
    
	if(def)
	{
	    initfunc_ << indent(1) << "rep_->" << name << " = "
		      << def << ";" << std::endl;
	}
    }
    return &contentele_;
}



ElementGenerator *ClassWriter::addOptionalProperty(const char *name,const char *id,
				      const char *typ,bool type_is_elementar)
{
    contentele_.checkfinished();
    contentele_.decl() << indent(level_+1) << "bool is_" << name << "_valid();"
		 << std::endl;
    contentele_.decl() << indent(level_+1) << "void set_" << name
		 << "_valid(bool to=true);"
		 << std::endl;
    if(!type_is_elementar || strcmp(typ,"String")==0)
    {
	body_ << "bool " << name_ << "::is_" << name << "_valid()\n{\n";
	body_ << indent(1) << "return (rep_->" << name << " != RAMS::RNIL);\n";
	body_ << "}\n" << std::endl;
	body_ << "void " << name_ << "::set_" << name << "_valid(bool to)\n{\n"
	      << indent(1) << "if(!to)\n" << indent(1) << "{\n"
	      << indent(2) << "rep_->" << name << " = RAMS::RNIL;\n"
	      << indent(1) << "}\n"
	      << "}\n" << std::endl;
    }
    else 
    {
	layout_ << indent(level_+2) << "bool " << name << "_valid;"
		<< std::endl;
	body_ << "bool " << name_ << "::is_" << name << "_valid()\n{\n";
	body_ << indent(1) << "return rep_->" << name << "_valid;\n";
	body_ << "}\n" << std::endl;
	body_ << "void " << name_ << "::set_" << name << "_valid(bool to)\n{\n"
	      << indent(1) << "rep_->" << name << "_valid = to;\n"
	      << "}\n" << std::endl;
	initfunc_ << indent(1) << "rep_->" << name << "_valid = false;"
		  << std::endl;
    }
    addProperty(name,id,typ,type_is_elementar,0);
    return &contentele_;
}

ElementGenerator *ClassWriter::addMultiProperty(const char *name,const char *id,
				   const char *typ,bool type_is_elementar)
{
    contentele_.checkfinished();
    if(type_is_elementar)
    {
	std::cerr << "Multiproperty for elementary typ " << typ
		  << " is not implemented " << std::endl;
	return 0;
    }
    if(!iteratorheaderincl_)
    {
	includes_ << "#include\"CommonListIterator.h\"" << std::endl;
    }
    includes_ << "#include\"" << typ << ".h\"" << std::endl;
    classpublic_ << indent(level_+1) << "/* id=" << id << "*/" << std::endl;
    layout_ << indent(level_+2) << "/* id=" << id << "*/" << std::endl;
    body_ << "/* id=" << id << "*/" << std::endl;
  
    layout_ << indent(level_+2) << "off_t " << name << "_root;" << std::endl;
    initfunc_ << indent(1) << "rep_->" << name << "_root=RAMS::RNIL;"
	      << std::endl;
    contentele_.decl() << indent(level_+1) << "class " << name << "_iterator "
		 << ": public RAMS::CommonListIterator\n"
		 << indent(level_+1) << "{\n"
		 << indent(level_+1) << "public:\n"
		 << indent(level_+2) <<  name << "_iterator(RAMS::GrowingBuffer& seg,off_t first);\n"
		 << indent(level_+2) << "~" <<  name << "_iterator();\n"
		 << indent(level_+2) << typ << "* operator()();\n"
		 << indent(level_+1) << "private:\n"
		 << indent(level_+2) << "RAMS::GrowingBuffer *gbuf_;\n"
		 << indent(level_+2) << typ << "* accessor_;\n"
		 << indent(level_+1) << "};\n"
		 << std::endl;

    contentele_.decl() << indent(level_+1) << name << "_iterator get_" << name << "_iterator();" << std::endl;
    body_ << name_ << "::" << name << "_iterator " << name_ << "::get_" << name << "_iterator()\n{\n"
          << indent(1) << "return " << name << "_iterator::" << name <<  "_iterator(*place_,rep_->"
	  << name << "_root);\n" 
	  << "}\n" << std::endl;

    body_ << name_ << "::" << name << "_iterator::" << name << "_iterator(RAMS::GrowingBuffer& seg,off_t first): RAMS::CommonListIterator(seg,first),gbuf_(&seg)\n{\n";
    body_ << indent(1) << "accessor_ = new " << typ << ";\n";
    body_ << "}\n\n";

    body_ << name_ << "::" << name << "_iterator::~" << name << "_iterator()\n"
	  << "{\n"
	  << indent(1) << "delete accessor_;\n"
	  << "}\n\n";
    body_ << typ << " *" <<  name_ << "::" << name
	  << "_iterator::operator()()\n"
	  << "{\n"
	  << indent(1) << "RAMS::octet_t *place = content();\n"
	  << indent(1) << "RAMS::MOF::typeid_t tid=*((::RAMS::MOF::typeid_t *)place);\n"
	  << indent(1) << "if(accessor_->_typeid()!=tid)\n"
	  << indent(1) << "{\n"
	  << indent(2) << "delete accessor_;\n"
	  << indent(2) << "accessor_ = dynamic_cast<" << typ
	  << " *>(resolve(tid));\n"
	  << indent(1) << "}\n"
	  << indent(1) << "accessor_->_setsegandrep(gbuf_,place+sizeof(::RAMS::MOF::typeid_t));\n"
	  << indent(1) << "return accessor_;\n"
	  << "}\n";

    classprivate_ << indent(level_+1) << "RAMS::CommonListAllocator " << name
		  << "_alloc_;" << std::endl;
      
    contentele_.decl() << indent(level_+1) << typ << "* add_" << name << "("
		       << typ << " *ta);" << std::endl;

    body_ << typ << " *" << name_ << "::add_" << name << "("
	  << typ << " *ta)\n{\n"
	  << indent(1) << "RAMS::octet_t *place = "
	  << name << "_alloc_.add(sizeof(::RAMS::MOF::typeid_t)+ta->_layout_size());\n"
	  << indent(1) << "*((RAMS::MOF::typeid_t *)place) = ta->_typeid();\n"
	  << indent(1) << "RAMS::octet_t *rep=place+sizeof(::RAMS::MOF::typeid_t);\n"
	  << indent(1) << "ta->_setsegandrep(place_,rep);\n"
	  << indent(1) << "ta->_init();\n"
	  << indent(1) << "return ta;\n"
	  << "}\n" << std::endl;

    contentele_.decl() << indent(level_+1) << "// " << typ << "& get_" << name 
		 << "(const Name& name);" << std::endl;
    constructorbody_ << indent(1) << name << "_alloc_.set(*place_,&(rep_->"
		     << name << "_root));" << std::endl;
    setsegbody_  << indent(1) << name << "_alloc_.set(*place_,&(rep_->"
		 << name << "_root));" << std::endl;
		
    return &contentele_;
}

void ClassWriter::addComment(const char *text,size_t len)
{
    for(const char *c=text;c<text+len;c++)
    {
	comment_ << *c;
    }
}

void ClassWriter::finish()
{
    Typename2Id typename2id;
    body_ << name_ << "::" << name_ << "():place_(0),rep_(0)"
	  << constructordecl_.str()
	  << "\n{\n}\n" << std::endl;
    body_ << initfunc_.str() << "}\n" <<  std::endl;

    body_ << "size_t " << name_ << "::_layout_size()\n{\n"
	  << indent(1) << "return sizeof(" << name_ << "::Layout);\n"
	  << "}\n\n";

    body_ << "::RAMS::MOF::typeid_t " << name_ << "::" << "_typeid()\n{\n"
	  << indent(1) << "return 0x" << std::hex
	  << typename2id(name_.c_str()) << ";"
	  << "\n}\n\n" << std::dec;

    body_ << name_ << "::~" << name_ << "()\n{\n"
	  << destructorbody_.str()
	  << "}\n";
    body_ << name_ << "::" << name_ << "(RAMS::GrowingBuffer *place):place_(0),rep_(0)"
	  << constructordecl_.str()
	  << "\n{\n"
	  << indent(1) << "RAMS::memchunk *mrep = (RAMS::memchunk *)(place->begin());\n"
	  << indent(1) << "place_=place;\n"
	  << indent(1) << "rep_=(Layout *)(mrep->mem);\n"
	  << indent(1) << "place->set_curlenptr(&(mrep->size));" << std::endl;
    body_ << constructorbody_.str() << std::endl;
    body_ << indent(1) << "if(place->length()==0)\n"
	  << indent(1) << "{\n"
	  << indent(2) << "place->wr_ptr(sizeof(size_t)+sizeof(Layout));\n"
	  << indent(2) << "_init();\n"
	  << indent(1) << "}\n"
	  << "}\n" << std::endl;
    
    setsegbody_	 << "}\n";
    body_ << setsegbody_.str() << std::endl;

    body_ << "void " << name_ << "::_callVisitor(::RAMS::MOF::Visitor *v)\n{\n";
    const char *sname = get_semantik_name();
    if(sname)
    {
	body_ << indent(1) << sname << "_Semantik *sem = dynamic_cast<"
	      << sname << "_Semantik *>(v);\n";
	body_ << indent(1) << "if(sem)\n" << indent(1) << "{\n";
	body_ << indent(2) << "sem->v" << name_ << "(this);\n";
	body_ << indent(1) << "}\n";
    }
    body_ << "}\n" << std::endl;
    body_.close();
    if(numbersuperclasses_==0)
    {
	classhead_ << ": public virtual RAMS::MOF::Object";
	includes_ << "#include<ramsmof/Object.h>" << std::endl;
    }
    header_ << includes_.str() << std::endl;
    if(namespace_)
    {
	header_ << "namespace " << namespace_ << "\n{\n";
    }
    header_ << indent(level_) << "/* " << comment_.str() << " */" << std::endl;
    header_ << classhead_.str() << std::endl;
    header_ << indent(level_) << "{\n"
	    << indent(level_) << "public:" << std::endl;
    header_ << indent(level_+1) << "struct Layout"  << "\n"
	    << indent(level_+1) << "{\n";
    header_ << layouthead_.str();
    header_ << layout_.str()  << indent(level_+1) << "};\n" << std::endl;
    header_ << indent(level_) << "public:" << std::endl;
    header_ << indent(level_+1) << name_
	    << "(RAMS::GrowingBuffer *place);" << std::endl;
    header_ << indent(level_+1) << name_ << "(); /* for instance variables*/\n"; 
    header_ << indent(level_+1) << "~" << name_ << "();\n";
    header_ << indent(level_+1) << "virtual void _setsegandrep(RAMS::GrowingBuffer *place,RAMS::octet_t *rep);\n";
    header_ << indent(level_+1) << "virtual size_t _layout_size();"
	    << std::endl;
    header_ << indent(level_+1) << "virtual ::RAMS::MOF::typeid_t _typeid();" << std::endl;
    header_ << indent(level_+1) << "virtual void _init();" << std::endl;
    header_ << indent(level_+1) << "virtual void _callVisitor(::RAMS::MOF::Visitor *v);" << std::endl;
    header_ << classpublic_.str() << std::endl;
   
    header_ << indent(level_) << "private:" << std::endl;
    header_ << indent(level_+1) << "RAMS::GrowingBuffer *place_;" << std::endl;
    header_ << indent(level_+1) << "Layout *rep_;" << std::endl;
    header_ << classprivate_.str();
    header_ << indent(level_) << "};" << std::endl;
    if(namespace_)
    {
	header_ << "};" << std::endl;
	free((void *)namespace_);
    }
    header_ << "#endif" << std::endl;
    header_.close();
}

ClassWriter::~ClassWriter()
{
    if(!finished_) finish();
}
