#include"classwriter.h"
#include<ctype.h>
#include"output.h"

void CommonContentElement::addComment(const char *text,size_t len)
{
    for(const char *c=text;c<text+len;c++)
    {
	comment_ << *c;
    }
}

void CommonContentElement::finish()
{
    std::string comment = comment_.str();
    if(comment.size()>0)
    {
	cw_->decl() << indent(cw_->level()+1) 
		    << "/* " << comment_.str() << " */" << std::endl;
    }
    cw_->decl() << decl_.str();
    comment_.str("");
    decl_.str("");
}

void CommonContentElement::checkfinished()
{
    if(!finished_) finish();
}

std::string toupper(std::string s)
{
    std::string ret(s.size(),' ');
    for(int i=0;i<s.length();i++)
    {
	ret[i]=toupper(s[i]);
    }
    return ret;
}

ClassWriter::ClassWriter(const char *namesp,const char *name,const char *id)
    :finished_(false),namespace_(0),name_(name),level_(0),
     numbersuperclasses_(0),iteratorheaderincl_(false),contentele_(this)
{
    std::string filename=std::string(name);
    if(namesp!=0)
    {
	namespace_=strdup(namesp);
	level_++;
    }
    std::string headerfilename = filename +".h";
    std::string bodyfilename = filename+".cc";
    header_.open(headerfilename.c_str(),std::ios_base::out);
    std::string inclstopper = toupper(filename)+"_H";
    header_ << "#ifndef " << inclstopper << std::endl;
    header_ << "#define " << inclstopper << std::endl;
    includes_ << "#include<RAMS/GrowingBuffer.h>" << std::endl;
    includes_ << "#include\"PrimitivTypes.h\"" << std::endl; 
    classhead_ << indent(level_) << "class " << name;
    

    body_.open(bodyfilename.c_str(),std::ios_base::out);
    body_ << "#include\"" << headerfilename << "\"\n" << std::endl;
    if(namesp !=0)
    {
	body_ << "using namespace " << namesp << ";\n" << std::endl;
    }
   
		 
    
    setsegbody_ << "void " << name_ << "::setsegandrep(RAMS::GrowingBuffer *place,Layout *rep)\n{\n"
	  << indent(1) << "place_=place;\n"
	  << indent(1) << "rep_=rep;\n";
   
    //??? what to do if place->freespace() < sizeof(Layout) ???
    initfunc_ << "void " << name_ << "::_init()\n{" << std::endl;
}

void ClassWriter::addSuperClass(const char *name)
{
    includes_ << "#include\"" << name << ".h\"" << std::endl;
    classhead_ << (numbersuperclasses_ ? ", " : " : ")
		 << "public " << name;
    layouthead_ << indent(level_+2) << name << "::Layout rep_" << name << ";\n";
    constructorbody_ << indent(1) << name 
		     << "::setsegandrep(place_,&(rep_->rep_" << name 
		     << "));" << std::endl;
    setsegbody_      << indent(1) << name 
		     << "::setsegandrep(place_,&(rep_->rep_" << name 
		     << "));" << std::endl;
    initfunc_   << indent(1) << name << "::_init();" << std::endl; 
    numbersuperclasses_++;
}

ElementGenerator *ClassWriter::addProperty(const char *name,const char *id,
			      const char *typ,bool typeelem,
			      const char *def)
{
    contentele_.checkfinished();
    classpublic_ << indent(level_+1) << "/* id=" << id << "*/" << std::endl;
    layout_ << indent(level_+2) << "/* id=" << id << "*/" << std::endl;
    body_ << "/* id=" << id << "*/" << std::endl;

    if(!typeelem)
    {
	includes_ << "#include\"" << typ << ".h\"" << std::endl;
	layout_ << indent(level_+2) << typ << "::Layout " << name << "_;"
		<< std::endl;
	contentele_.decl() << indent(level_+1) << typ << " " << name << "_;"
		     << std::endl;
	constructorbody_ << indent(1) << name
			 << "_.setsegandrep(place_,&(rep_->" << name << "_));"
			 << std::endl;
	setsegbody_      << indent(1) << name
			 << "_.setsegandrep(place_,&(rep_->" << name << "_));"
			 << std::endl;
	initfunc_ << indent(1) <<  name << "_._init();\n";
    }
    else if(strcmp(typ,"String")==0)
    {
	contentele_.decl() << indent(level_+1) << "const RAMS::String &"
		     << name << "();" << std::endl;
	contentele_.decl() << indent(level_+1) << "void " << name
		     << "(const char *value);\n" << std::endl;
	layout_ << indent(level_+2) << "off_t " << name << ";" << std::endl;
	body_ << "const RAMS::String & " << name_ << "::" << name << "()\n"
	      << "{\n"
	      << indent(1) << "return *((RAMS::String *)((octet_t *)(place_->begin())+rep_->"
	      << name << "));\n"
	      << "}\n\n" << std::endl;
	body_ << "void " << name_ << "::" << name << "(const char *value)\n" 
	      << "{\n";	
	body_ << indent(1) << "RAMS::String *z;\n"
	      << indent(1) << "const char *p=value;\n"
	      << indent(1) << "place_->wrptr_align(2);\n"
	      << indent(1) << "rep_->" << name << " = place_->wr_rel();\n"
	      << indent(1) << "z=(RAMS::String *)(place_->wr_ptr(sizeof(size_t)));\n"
	      << indent(1) << "z->size=0;\n"
	      << indent(1) << "while(*p!='\\0')\n"
	      << indent(1) << "{\n"
	      << indent(2) << "*(place_->wr_ptr(1))=*p;\n"
	      << indent(2) << "p++; (z->size)++;\n"
	      << indent(1) << "}\n"
	      << "}\n" << std::endl;
	initfunc_ << indent(1) << "rep_->" << name << " = RAMS::RNIL;"
		  << std::endl;
	if(def)
	{
	    initfunc_ << indent(1) << name << "(\"" << def
		      << "\");" << std::endl;
	}
    }
    else
    {
        contentele_.decl() << indent(level_+1) << typ << " " 
			   << name << "();" << std::endl;
	contentele_.decl() << indent(level_+1) << "void " << name << "("
			   << typ << " value);" << std::endl;
    
	layout_ << indent(level_+2) << typ << " " << name << ";" << std::endl;
	body_ << typ << " " << name_ << "::" << name << "()\n"
	      << "{\n"
	      << indent(1) << "return rep_->" << name << ";\n"
	      << "}\n" << std::endl;
	body_ << "void " << name_ << "::" << name << "(" << typ << " value)\n" 
	      << "{\n"
	      << indent(1) << "rep_->" << name << " = value;\n"
	      << "}\n" << std::endl;
    
	if(def)
	{
	    initfunc_ << indent(1) << "rep_->" << name << " = "
		      << def << ";" << std::endl;
	}
    }
    return &contentele_;
}



ElementGenerator *ClassWriter::addOptionalProperty(const char *name,const char *id,
				      const char *typ,bool typeelem)
{
    contentele_.checkfinished();
    contentele_.decl() << indent(level_+1) << "bool is_" << name << "_valid();"
		 << std::endl;
    contentele_.decl() << indent(level_+1) << "void set_" << name
		 << "_valid(bool to=true);"
		 << std::endl;
    if(strcmp(typ,"String")==0)
    {
	body_ << "bool " << name_ << "::is_" << name << "_valid()\n{\n";
	body_ << indent(1) << "return (rep_->" << name << " != RAMS::RNIL);\n";
	body_ << "}\n" << std::endl;
	body_ << "void " << name_ << "::set_" << name << "_valid(bool to)\n{\n"
	      << indent(1) << "if(!to)\n" << indent(1) << "{\n"
	      << indent(2) << "rep_->" << name << " = RAMS::RNIL;\n"
	      << indent(1) << "}\n"
	      << "}\n" << std::endl;
    }
    else 
    {
	layout_ << indent(level_+2) << "bool " << name << "_valid;"
		<< std::endl;
	body_ << "bool " << name_ << "::is_" << name << "_valid()\n{\n";
	body_ << indent(1) << "return rep_->" << name << "_valid;\n";
	body_ << "}\n" << std::endl;
	body_ << "void " << name_ << "::set_" << name << "_valid(bool to)\n{\n"
	      << indent(1) << "rep_->" << name << "_valid = to;\n"
	      << "}\n" << std::endl;
	initfunc_ << indent(1) << "rep_->" << name << "_valid = false;"
		  << std::endl;
    }
    addProperty(name,id,typ,typeelem,0);
    return &contentele_;
}

ElementGenerator *ClassWriter::addMultiProperty(const char *name,const char *id,
				   const char *typ,bool typeelem)
{
    contentele_.checkfinished();
    if(typeelem)
    {
	std::cerr << "Multiproperty for elementary typ " << typ
		  << " is not implemented " << std::endl;
	return 0;
    }
    if(!iteratorheaderincl_)
    {
	includes_ << "#include\"CommonListIterator.h\"" << std::endl;
    }
    includes_ << "#include\"" << typ << ".h\"" << std::endl;
    classpublic_ << indent(level_+1) << "/* id=" << id << "*/" << std::endl;
    layout_ << indent(level_+2) << "/* id=" << id << "*/" << std::endl;
    body_ << "/* id=" << id << "*/" << std::endl;
  
    layout_ << indent(level_+2) << "off_t " << name << "_root;" << std::endl;
    initfunc_ << indent(1) << "rep_->" << name << "_root=RAMS::RNIL;"
	      << std::endl;
    contentele_.decl() << indent(level_+1) << "class " << name << "_iterator "
		 << ": public RAMS::CommonListIterator\n"
		 << indent(level_+1) << "{\n"
		 << indent(level_+1) << "public:\n"
		 << indent(level_+2) <<  name << "_iterator(RAMS::GrowingBuffer& seg,off_t first);\n"
		 << indent(level_+2) << "~" <<  name << "_iterator();\n"
		 << indent(level_+2) << typ << "& operator()();\n"
		 << indent(level_+1) << "private:\n"
		 << indent(level_+2) << "RAMS::GrowingBuffer *gbuf_;\n"
		 << indent(level_+2) << typ << "* accessor_;\n"
		 << indent(level_+1) << "};\n"
		 << std::endl;

    contentele_.decl() << indent(level_+1) << name << "_iterator get_" << name << "_iterator();" << std::endl;
    body_ << name_ << "::" << name << "_iterator " << name_ << "::get_" << name << "_iterator()\n{\n"
          << indent(1) << "return " << name << "_iterator::" << name <<  "_iterator(*place_,rep_->"
	  << name << "_root);\n" 
	  << "}\n" << std::endl;

    body_ << name_ << "::" << name << "_iterator::" << name << "_iterator(RAMS::GrowingBuffer& seg,off_t first): RAMS::CommonListIterator(seg,first),gbuf_(&seg)\n{\n";
    body_ << indent(1) << "accessor_ = new " << typ << ";\n";
    body_ << "}\n\n";

    body_ << name_ << "::" << name << "_iterator::~" << name << "_iterator()\n"
	  << "{\n"
	  << indent(1) << "delete accessor_;\n"
	  << "}\n\n";
    body_ << typ << "& " <<  name_ << "::" << name
	  << "_iterator::operator()()\n"
	  << "{\n"
	  << indent(1) << "accessor_->setsegandrep(gbuf_,(" << typ << "::Layout *)(content()));\n"
	  << indent(1) << "return *accessor_;\n"
	  << "}\n";

    classprivate_ << indent(level_+1) << "RAMS::CommonListAllocator " << name
		  << "_alloc_;" << std::endl;
    classprivate_ << indent(level_+1) << typ << " *" << name << "_;"
	          << " // Accessor for adding a value "
		  << std::endl;
      
    contentele_.decl() << indent(level_+1) << typ << "& add_" << name << "();" << std::endl;
    constructordecl_ << "," << name << "_(0)";
    body_ << typ << "& " << name_ << "::add_" << name << "()\n{\n"
	  << indent(1) << "if(!" << name << "_)\n"
	  << indent(1) << "{\n"
	  << indent(2) << name << "_ = new " << typ << ";\n"
	  << indent(1) << "}\n"
	  << indent(1) << typ << "::Layout *rep = (" << typ << "::Layout *)"
	  << name << "_alloc_.add(sizeof(" << typ << "::Layout));\n"
	  << indent(1) << name << "_->setsegandrep(place_,rep);\n"
	  << indent(1) << name << "_->_init();\n"
	  << indent(1) << "return *" << name << "_;\n"
	  << "}\n" << std::endl;

    destructorbody_ << indent(1) << "if(" << name << "_) delete "
		    << name << "_;" << std::endl;

    contentele_.decl() << indent(level_+1) << "// " << typ << "& get_" << name 
		 << "(const Name& name);" << std::endl;
    constructorbody_ << indent(1) << name << "_alloc_.set(*place_,&(rep_->"
		     << name << "_root));" << std::endl;
    setsegbody_  << indent(1) << name << "_alloc_.set(*place_,&(rep_->"
		 << name << "_root));" << std::endl;
		
    return &contentele_;
}

void ClassWriter::addComment(const char *text,size_t len)
{
    for(const char *c=text;c<text+len;c++)
    {
	comment_ << *c;
    }
}

void ClassWriter::finish()
{
    body_ << name_ << "::" << name_ << "():place_(0),rep_(0)"
	  << constructordecl_.str()
	  << "\n{\n}\n" << std::endl;
    body_ << initfunc_.str() << "}\n" <<  std::endl;
    body_ << name_ << "::~" << name_ << "()\n{\n"
	  << destructorbody_.str()
	  << "}\n";
    body_ << name_ << "::" << name_ << "(RAMS::GrowingBuffer *place):place_(0),rep_(0)"
	  << constructordecl_.str()
	  << "\n{\n"
	  << indent(1) << "RAMS::memchunk *mrep = (RAMS::memchunk *)(place->begin());\n"
	  << indent(1) << "place_=place;\n"
	  << indent(1) << "rep_=(Layout *)(mrep->mem);\n"
	  << indent(1) << "place->set_curlenptr(&(mrep->size));" << std::endl;
    body_ << constructorbody_.str() << std::endl;
    body_ << indent(1) << "if(place->length()==0)\n"
	  << indent(1) << "{\n"
	  << indent(2) << "place->wr_ptr(sizeof(size_t)+sizeof(Layout));\n"
	  << indent(2) << "_init();\n"
	  << indent(1) << "}\n"
	  << "}\n" << std::endl;
    
    setsegbody_	 << "}\n";
    body_ << setsegbody_.str() << std::endl;
    body_.close();
    header_ << includes_.str() << std::endl;
    if(namespace_)
    {
	header_ << "namespace " << namespace_ << "\n{\n";
    }
    header_ << indent(level_) << "/* " << comment_.str() << " */" << std::endl;
    header_ << classhead_.str() << std::endl;
    header_ << indent(level_) << "{\n"
	    << indent(level_) << "public:" << std::endl;
    header_ << indent(level_+1) << "struct Layout"  << "\n"
	    << indent(level_+1) << "{\n";
    header_ << layouthead_.str();
    header_ << layout_.str()  << indent(level_+1) << "};\n" << std::endl;
    header_ << indent(level_) << "public:" << std::endl;
    header_ << indent(level_+1) << name_
	    << "(RAMS::GrowingBuffer *place);" << std::endl;
    header_ << indent(level_+1) << "void _init();" << std::endl;
    header_ << classpublic_.str() << std::endl;
    header_ << indent(level_+1) << "/* for internal use: instance variables*/\n";
    header_ << indent(level_+1) << name_ << "();\n";
    header_ << indent(level_+1) << "~" << name_ << "();\n";
    header_ << indent(level_+1) << "void setsegandrep(RAMS::GrowingBuffer *place,Layout *rep);\n";
    header_ << indent(level_) << "private:" << std::endl;
    header_ << indent(level_+1) << "RAMS::GrowingBuffer *place_;" << std::endl;
    header_ << indent(level_+1) << "Layout *rep_;" << std::endl;
    header_ << classprivate_.str();
    header_ << indent(level_) << "};" << std::endl;
    if(namespace_)
    {
	header_ << "};" << std::endl;
	free((void *)namespace_);
    }
    header_ << "#endif" << std::endl;
    header_.close();
}

ClassWriter::~ClassWriter()
{
    if(!finished_) finish();
}
