#include"L0filler.h"
#include"L0ElementGenerator.h"
#include"ModelError.h"

extern "C" void set_ramsomf_factory(AbstractCodeFactory *f);
extern "C" const char *get_semantik_name();
class Global
{
public:
  Global()
  {
     set_ramsomf_factory(&l0factory);
  };
private:
  L0FillerFactory l0factory;
};

Global global;
static  L0ModelWriter model(4);

L0FillerFactory::L0FillerFactory()
{
    model.open(get_semantik_name());
}

L0FillerFactory::~L0FillerFactory()
{
    model.save();
}


PackageGenerator *L0FillerFactory::genPackage(Scope *scope,const char *name,const char *id,const char *uri)
{
    return new L0PackageFiller(scope,name,id,uri);
}



ClassGenerator *L0FillerFactory::genClass(Scope *scope,const char *name,const char *id)
{
    return new L0ClassFiller(scope,name,id);
}

L0PackageFiller::L0PackageFiller(Scope *scope,const char *name,const char *id,const char *uri):name_(name)
{
    if(scope==0)
    {
	if(model.root()==0)
	{
	    model.root(&package_);
	}
	else
	{
	    throw ModelError("Only one Root-Element allowed");
	}
    }
    else 
    {
	L0PackageFiller *nested=dynamic_cast<L0PackageFiller *>(scope);
	if(nested==0)
	{
	    throw ModelError("Scope %s is not a Package",scope->Name().c_str());
	}
	nested->package()->add_nestedPackage(&package_);
    }
    package_.name(name);
}

std::string L0PackageFiller::Name()
{
    return name_;
}

void L0PackageFiller::addComment(const char *text,size_t len)
{
    L0::Comment acomment;
    package_.add_ownedComment(&acomment);
    acomment.body(text,len);
}

void L0PackageFiller::finish()
{
}

L0ClassFiller::L0ClassFiller(Scope *scope,const char *name,const char *id):name_(name)
{
    if(scope==0)
    {
	if(model.root()==0)
	{
	    model.root(&class_);
	}
	else
	{
	    throw ModelError("Only one Root-Element allowed");
	}
    }
    else 
    {
	L0PackageFiller *nested=dynamic_cast<L0PackageFiller *>(scope);
	if(nested==0)
	{
	    throw ModelError("Scope %s is not a Package",scope->Name().c_str());
	}
	nested->package()->add_ownedType(&class_);
    }
    class_.name(name);
}

std::string L0ClassFiller::Name()
{
    return name_;
}

void L0ClassFiller::addComment(const char *text,size_t len)
{
    L0::Comment acomment;
    class_.add_ownedComment(&acomment);
    acomment.body(text,len);
}

void L0ClassFiller::finish()
{
}

void 
L0ClassFiller::addSuperClass(const char *name)
{
}

ElementGenerator *
L0ClassFiller::addProperty(const char *name,const char *id,
			   const char *typ,bool typeelem,
			   const char *def)
{
    L0::Property *property = new L0::Property();
    L0ElementGenerator *ret = new L0ElementGenerator(property);
    class_.add_ownedAttribute(property);
    property->name(name);
    if(def)
    {
	property->default_(def);
    }
    property->set_lower_valid(true);
    property->lower(1);
    property->set_upper_valid(true);
    property->upper(1);

    // How to set the type ?
    // - it must be resolved and a reference must be set
    return ret;
}

ElementGenerator *
L0ClassFiller::addMultiProperty(const char *name,const char *id,
				const char *typ,bool typeelem)
{
    L0::Property *property = new L0::Property();
    L0ElementGenerator *ret = new L0ElementGenerator(property);
    class_.add_ownedAttribute(property);
    property->name(name);
    property->set_lower_valid(true);
    property->lower(0);
    // How to set the type ?
    // - it must be resolved and a reference must be set
    return ret;
}

ElementGenerator *
L0ClassFiller::addOptionalProperty(const char *name,
				   const char *id,
				   const char *typ,
				   bool typeelem)
{
    L0::Property *property = new L0::Property();
    L0ElementGenerator *ret = new L0ElementGenerator(property);
    class_.add_ownedAttribute(property);
    property->name(name);
    property->set_lower_valid(true);
    property->lower(0);
    property->set_upper_valid(true);
    property->upper(1);
    // How to set the type ?
    // - it must be resolved and a reference must be set
    return ret;
}
    
L0ClassFiller::~L0ClassFiller()
{
}
