#include<iostream>
#include<libxml/xmlmemory.h>
#include<libxml/parser.h>
#include"coutput.h"
#include"XMIParser_AttributInterpreter.h"
#include"XMIParserContext_Class.h"
#include"XMIParserContext_Comment.h"
#include"XMIParserContext_Operation.h"
#include"XMIParserContext_Rule.h"
#include<alloca.h>
using namespace XMIParser;

bool isType_elementary(const char *type_name)
{
    if(strcmp("Boolean",type_name)==0
       || strcmp("Integer",type_name)==0
       || strcmp("String",type_name)==0
       || strcmp("UnlimitedNatural",type_name)==0
	)
    {
	return true;
    }
    return false;
}

ParserContext_Class::ParserContext_Class(XMIparser *parser,const char *name,const char *xmiid,const AttributTab &atts):ParserContext(parser),classgen_(0),currele_(0)
{
    const char *namesp=0;
    Scope *upperscope=0;
    name_= std::string(name);
    
    if(parser->get_deepth()> 2)
    {
	namesp=parser->top()->get_Name().c_str();
	upperscope = parser->top()->get_Scope();
    }
    classgen_ = the_factory->genClass(upperscope,name,xmiid);
    currele_=classgen_;
    std::string sckey("superClass");
    AttributTab::const_iterator sc = atts.find(sckey);
    if(sc != atts.end())
    {
	std::string superclass_str = sc->second;
	const char *superclass=superclass_str.c_str();
	char *buffer=(char *)alloca(strlen(superclass)+1);
	char *p=buffer;
	char *sc;
        strcpy(buffer,superclass);
	while((sc=strsep(&p," \n"))!=NULL)
	{
	    classgen_->addSuperClass(sc);
	}
    }
}

ParserContext_Class::~ParserContext_Class()
{
    delete classgen_;
}

std::string ParserContext_Class::get_Name()
{
    return name_;
}

void ParserContext_Class::startelement(const xmlChar *name,const xmlChar **atts)
{
    inclevel();
    
    if(xmlStrcmp(name, (const xmlChar *) "ownedComment")==0)
    {
	ParserContext_Comment *comment=new ParserContext_Comment(parser_,atts);
	level_--;
	parser_->push(comment);
    }
    else if(xmlStrcmp(name, (const xmlChar *) "ownedAttribute")==0)
    {
	std::cerr << "(" << level_ << ") Attribute " << std::endl;
	if(atts)
	{
	    AttributInterpreter ai; 
	    ai.parse(atts);
	    AttributTab& at = *(ai.getTab());
	    if(strcmp(ai.xmityp(),"cmof:Property")==0)
	    {
		bool iscomposite=false;
		AttributTab::const_iterator ci = at.find(std::string("isComposite"));
		iscomposite=(ci!=at.end() && ci->second == "true");
		AttributTab::const_iterator cia = at.find(std::string("association"));

		if((ai.lower()==0 || strcmp(ai.lower(),"1")==0)
		   && (ai.upper()==0 || strcmp(ai.upper(),"1")==0))
		{
		    const char *def=0;
		    AttributTab::const_iterator ci = at.find(std::string("default"));
		    if(ci!=at.end())
		    {
			def=ci->second.c_str();
		    }
		    currele_=classgen_->addProperty(ai.name(),ai.xmiid(),
					   ai.typ(),isType_elementary(ai.typ()),
					   def);
		} 
		else if((ai.lower()!=0 && strcmp(ai.lower(),"0")==0)
			&& (ai.upper()==0 || strcmp(ai.upper(),"1")==0))
		{
		    if(iscomposite || isType_elementary(ai.typ()))
		    {
			currele_=classgen_->addOptionalProperty(ai.name(),ai.xmiid(),
								ai.typ(),isType_elementary(ai.typ()));
		    }
		    else
		    {
			currele_=0;
		    	std::cerr << "Association end " //<< ai.name()
		    		  << std::endl;
		    }
		}
		else if(iscomposite
			&& (ai.lower()!=0 || ai.upper()!=0))
		{
		    currele_=classgen_->addMultiProperty(ai.name(),ai.xmiid(),
						ai.typ(),isType_elementary(ai.typ()));
		}
		else
		{
		    currele_=0;
		    std::cerr << "unhandeled multplicity: lower= "
			      << (ai.lower() ? ai.lower() :  "NULL")
			      << " upper=" << (ai.upper() ? ai.upper() :  "NULL")
			      << " iscomposite="
			      << (iscomposite ? "true" : "false")
			      << std::endl;
		    if(cia != at.end())
		    {
			std::cerr << "Association: " << cia->second << std::endl;
		    }
		}
	    }
	    else
	    {
		std::cerr << "Unkown Attribute " << name 
			  << " of type " << ai.xmityp() << std::endl;
	    }
	}
	else
	{
	    std::cerr << "OwnedAttribute without Attributes " << std::endl;
	}
    }
    else if(xmlStrcmp(name, (const xmlChar *) "ownedOperation")==0)
    {
	std::cerr << "(" << level_ << ") Operation " << std::endl;
	currele_=0;
	if(atts)
	{
	    AttributInterpreter ai; 
	    ai.parse(atts);
	    if(strcmp(ai.xmityp(),"cmof:Operation")==0)
	    {
		level_--;
		ParserContext_Operation *op=new ParserContext_Operation(parser_,ai.name(),ai.xmiid());
		parser_->push(op);
	    }
	    else
	    {
		std::cerr << "Operation with xmi::type=" << ai.xmityp()
			  << std::endl;
	    }
	}
	else
	{
	    std::cerr << "Operation without attributs " << std::endl;
	}
    }
    else if(xmlStrcmp(name, (const xmlChar *) "ownedRule")==0)
    { // for attribute
	currele_=0;
	ParserContext_Rule *rule = new ParserContext_Rule(parser_);
	level_--;
	parser_->push(rule);
    }
    else
    {
	std::cerr << "(" << level_ << ")";
	std::cerr << "Class:startelement " << name << std::endl;
    }
}

void ParserContext_Class::endelement(const xmlChar *name)
{
    std::cerr << "(" << level_ << ") Class: endelement " << name << std::endl;
    if((xmlStrcmp(name, (const xmlChar *) "ownedAttribute")==0
	|| xmlStrcmp(name, (const xmlChar *) "ownedOperation")==0)
       && currele_)
    {
	currele_->finish();
    }
    declevel();
}

void ParserContext_Class::addComment(const xmlChar *value,int len)
{
    if(currele_)
    {
	currele_->addComment((const char *)value,len);
    }
}
