#include<iostream>
#include<stdarg.h>
#include"xmisax.h"
#include"XMIParserContext_Initial.h"
#include"ClassGenerator.h"
#include<dlfcn.h>

AbstractCodeFactory *the_factory;
static char *semantikname = 0;
extern "C" void set_ramsomf_factory(AbstractCodeFactory *f)
{
  the_factory=f;
}

extern "C" const char *get_semantik_name()
{
    return semantikname;
}

using namespace XMIParser;
const char *XMLError::what() const throw()
{
    return "XML Parsing Error";
}

XMIError::XMIError(const char *msg, ...)
{
    va_list ap;
    va_start(ap, msg);
    vsnprintf(msg_,256,msg,ap);
    va_end(ap);
}

const char *XMIError::what() const throw()
{
    return msg_;
}

void XMIparser::startDocument(void *user_data)
{
    std::cerr << "startDocument" << std::endl;
    XMIparser *self = (XMIparser *)user_data;
    //self->cur_context_=new ParserContext_Initial(self);
    self->push(new ParserContext_Initial(self));
};

void XMIparser::endDocument(void *userdata)
{
    std::cerr << "endDocument" << std::endl;
}

void XMIparser::startelement(void *ctx,const xmlChar *name,const xmlChar **atts)
{
    XMIparser *self = (XMIparser *)ctx;
    //std::cerr << "#StartElement " << name << std::endl;
    //if(atts!=0)
    //{
    //for(const xmlChar **a=atts;*a!=0;a++)
    //{
    //    std::cerr << "\t" << *a << std::endl;
    //}
    //}
    self->cur_context_->startelement(name,atts);
}

void XMIparser::endelement(void *ctx,const xmlChar *name)
{
    XMIparser *self = (XMIparser *)ctx;
    //std::cerr << "#EndElement " << name << std::endl;
    self->cur_context_->endelement(name);
}

void XMIparser::attribute(void *ctx,const xmlChar *name,const xmlChar *value)
{
    std::cerr << "\tAttribute " << name << "=" << value << std::endl;
}

void XMIparser::characters(void *ctx,const xmlChar *value,int len)
{
    XMIparser *self = (XMIparser *)ctx;
    self->cur_context_->characters(value,len);
}

void XMIparser::warning(void *ctx,const char *msg, ...)
{
    va_list ap;
    va_start(ap, msg);
    fprintf(stderr,"WARNING:");vfprintf(stderr,msg,ap);
    va_end(ap);
}

void XMIparser::error(void *ctx,const char *msg, ...)
{
    va_list ap;
    va_start(ap, msg);
    fprintf(stderr,"ERROR:");vfprintf(stderr,msg,ap);
    va_end(ap);
}

void XMIparser::fatalerror(void *ctx,const char *msg, ...)
{
    va_list ap;
    va_start(ap, msg);
    fprintf(stderr,"FATAL ERROR:");vfprintf(stderr,msg,ap);
    va_end(ap);
    throw XMLError();
}

XMIparser::XMIparser():cur_context_(0),deepth_(0)
{
   handler.internalSubset=NULL;
   handler.isStandalone=NULL;
   handler.hasInternalSubset=NULL;
   handler.hasExternalSubset=NULL;
   handler.resolveEntity=NULL;
   handler.getEntity=NULL;
   handler.entityDecl=NULL;
   handler.notationDecl=NULL;
   handler.attributeDecl=NULL;
   handler.elementDecl=NULL;
   handler.unparsedEntityDecl=NULL;
   handler.setDocumentLocator=NULL;
   handler.startDocument=startDocument;
   handler.endDocument=endDocument;
   handler.startElement=startelement;
   handler.endElement=endelement;
   // no attribute ???
   handler.reference=NULL;
   handler.characters=characters;
   handler.ignorableWhitespace=NULL;
   handler.processingInstruction=NULL;
   handler.comment=NULL;
   handler.warning=XMIparser::warning;
   handler.error=XMIparser::error;
   handler.fatalError=fatalerror;
   handler.getParameterEntity=NULL;
   handler.cdataBlock=NULL;
   handler.externalSubset=NULL;
   handler.initialized=1;
}

void XMIparser::parse(const char *filename)
{
    int r=xmlSAXUserParseFile(&handler,this,filename);
    if(r<0) throw XMLError();
}

ParserContext *XMIparser::top()
{
    return cur_context_;
}

void XMIparser::push(ParserContext *ctx)
{
    ctx->prev_=cur_context_;
    cur_context_=ctx;
    deepth_++;
}

void XMIparser::pop()
{
    ParserContext *cur = cur_context_;
    if(cur_context_)
    {
	cur_context_ = cur_context_->prev_;
	delete cur;
	deepth_--;
    }
}

int XMIparser::get_deepth()
{
    return deepth_;
}

int main(int argc,const char **argv)
{
    const char *semantikso="codewriter.so";
    XMIparser xmip;
    const char *docname=0;
    const char **arg=argv+1;
    while(*arg!=0 && **arg=='-')
    {
	if(strcmp(*arg,"-s")==0)
	{
	    semantikso=*(++arg);
	}
	arg++;
    }
    if (*arg==0)
    {
	std::cerr << "Usage:" << argv[0] << "docname" << std::endl;
	return 1;
    }
    docname=*arg;
    std::string docnamestr = basename(docname);
    std::string semname=docnamestr.substr(0,docnamestr.rfind("."));
    std::cerr << "Semantikname " << semname << std::endl;
    semantikname = strdup(semname.c_str());
    void *dlh = dlopen(semantikso,RTLD_NOW);	
    if(dlh==0)
    {
	std::cerr << "Loading " << semantikso << ":"
		  << dlerror() << std::endl;
        exit(1);
    }
    if(the_factory==0)
    {
	std::cerr << "No factory found " << std::endl;
	exit(1);
    }
    try
    {
	xmip.parse(docname);
    }
    catch(XMIError& ex)
    {
	std::cerr << ex.what() << std::endl;
    }
    catch(std::exception &ex)
    {
	std::cerr << ex.what() << std::endl;
    }
    dlclose(dlh);
}
